# Generated by Django 5.2.4 on 2025-10-20 13:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('history4feed', '0005_feed_datetime_modified'),
    ]

    operations = [
        migrations.AddField(
            model_name='job',
            name='completion_time',
            field=models.DateTimeField(default=None, null=True),
        ),
    ]



from datetime import UTC, datetime
import uuid
import pytest
from history4feed.app.models import Feed, Job, Post


@pytest.fixture
def feeds():
    return [
        Feed.objects.create(
            url="https://example.com/rss1.xml",
            title="Test Feed 1",
            feed_type="atom",
            description="some description",
            id=uuid.UUID("6ca6ce37-1c69-4a81-8490-89c91b57e557"),
        ),
        Feed.objects.create(
            url="https://example.com/rss2.xml",
            title="Test Feed 2",
            feed_type="rss",
            description="descr-iption",
            id=uuid.UUID("0dfccb58-158c-4436-b338-163e3662943c"),
        ),
        Feed.objects.create(
            url="https://example.com/rss3.xml",
            title="Some other feed 3",
            feed_type="skeleton",
        ),
    ]


@pytest.fixture
def feed(feeds) -> Feed:
    return feeds[0]


@pytest.fixture
def feed_posts(feed):
    p1 = Post.objects.create(
        feed=feed,
        title="First post",
        pubdate=datetime.now(UTC),
        link="https://example.net/post1",
        id=uuid.UUID("561ed102-7584-4b7d-a302-43d4bca5605b"),
    )
    p2 = Post.objects.create(
        feed=feed,
        title="Second post",
        pubdate=datetime.now(UTC),
        link="https://example.net/post2",
    )
    return feed, (p1, p2)


@pytest.fixture
def jobs(feeds):
    return [
        Job.objects.create(feed=feeds[0], id="8ff3672d-067b-40af-9065-e801061f5593"),
        Job.objects.create(feed=feeds[1], id="e9794a6c-388e-4bd5-bf29-6bc01aebb8bb"),
    ]



@pytest.fixture(scope='session')
def api_schema():
    import schemathesis
    from history4feed.asgi import application
    yield schemathesis.openapi.from_asgi("/api/schema/?format=json", application)


from drf_spectacular.openapi import AutoSchema
from rest_framework.serializers import Serializer
from rest_framework.views import exception_handler
from rest_framework.exceptions import ValidationError
from django.core import exceptions
from dogesec_commons.utils.autoschema import CustomAutoSchema

class H4FSchema(CustomAutoSchema):
    def _is_list_view(self, serializer: Serializer | type[Serializer] | None = None) -> bool:
        if self.path.endswith("/xml/"):
            return True
        return super()._is_list_view(serializer)


from rest_framework import pagination, response, renderers
from rest_framework.filters import OrderingFilter, BaseFilterBackend
from django.utils.encoding import force_str
from django.db.models import Q
from datetime import datetime, UTC
import typing
from dogesec_commons.utils import Pagination, Ordering
from dogesec_commons.utils.filters import MinMaxDateFilter, DatetimeFilter
from django.utils import timezone
from django.forms import DateTimeField
from django_filters.rest_framework import filters

# use pagination to modify how xml/rss renders
class XMLPostPagination(Pagination):
    def get_paginated_response_schema(self, schema):
        return {
            'type': 'string',
            'example': '<?xml version="1.0" ?>'
        }
    
    def get_paginated_response(self, data):
        return response.Response(data, headers={
            'rss_page_size': self.get_page_size(self.request),
            'rss_page_number': self.page.number,
            'rss_page_results_count': len(self.page),
            'rss_total_results_count': self.page.paginator.count,
        }, content_type="application/rss+xml; charset=UTF-8")
    
    def get_schema_operation_parameters(self, view):
        return super().get_schema_operation_parameters(view)

class RSSRenderer(renderers.BaseRenderer):
    media_type = "application/rss+xml"
    format = "xml"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        return data


import pytest
from unittest.mock import patch, MagicMock
from requests import Response
from history4feed.app.settings import History4FeedServerSettings
from history4feed.h4fscripts.h4f import (
    fetch_page_with_retries,
    fetch_page,
    fetch_with_scapfly,
    FatalError,
    parse_feed_from_url
)
from history4feed.h4fscripts.exceptions import history4feedException, ScrapflyError, FetchRedirect
from types import SimpleNamespace
from history4feed.h4fscripts.h4f import get_full_text
from history4feed.h4fscripts.exceptions import history4feedException
from readability import Document as ReadabilityDocument

@pytest.fixture
def dummy_url():
    return "https://example.com/test"


@patch("history4feed.h4fscripts.h4f.time.sleep", return_value=None)
@patch("history4feed.h4fscripts.h4f.fetch_page")
@patch("history4feed.h4fscripts.h4f.fake_useragent.UserAgent")
def test_fetch_page_with_retries_success(mock_ua, mock_fetch_page, mock_sleep, dummy_url):
    mock_ua().random = "test-agent"
    mock_fetch_page.return_value = (b"html", "text/html", dummy_url)

    result = fetch_page_with_retries(dummy_url)
    assert result == (b"html", "text/html", dummy_url)
    assert mock_fetch_page.call_count == 1
    mock_sleep.assert_not_called()


@patch("history4feed.h4fscripts.h4f.time.sleep", return_value=None)
@patch("history4feed.h4fscripts.h4f.fetch_page")
@patch("history4feed.h4fscripts.h4f.fake_useragent.UserAgent")
def test_fetch_page_with_retries_with_retry(mock_ua, mock_fetch_page, mock_sleep, dummy_url):
    mock_ua().random = "test-agent"
    mock_fetch_page.side_effect = [Exception("fail"), (b"ok", "text/html", dummy_url)]

    result = fetch_page_with_retries(dummy_url, retry_count=1)
    assert result == (b"ok", "text/html", dummy_url)
    assert mock_fetch_page.call_count == 2
    mock_sleep.assert_called_once()


@patch("history4feed.h4fscripts.h4f.time.sleep", return_value=None)
@patch("history4feed.h4fscripts.h4f.fetch_page")
@patch("history4feed.h4fscripts.h4f.fake_useragent.UserAgent")
def test_fetch_page_with_retries_fatal_error(mock_ua, mock_fetch_page, mock_sleep, dummy_url):
    mock_ua().random = "test-agent"
    mock_fetch_page.side_effect = FatalError("fatal")

    with pytest.raises(FatalError):
        fetch_page_with_retries(dummy_url)
    mock_sleep.assert_not_called()


@patch("history4feed.h4fscripts.h4f.time.sleep", return_value=None)
@patch("history4feed.h4fscripts.h4f.fetch_page")
@patch("history4feed.h4fscripts.h4f.fake_useragent.UserAgent")
def test_fetch_page_with_retries_exhausts(mock_ua, mock_fetch_page, mock_sleep, dummy_url):
    mock_ua().random = "test-agent"
    mock_fetch_page.side_effect = Exception("fail")

    with pytest.raises(ConnectionError) as e:
        fetch_page_with_retries(dummy_url, retry_count=2)
    assert "could not fetch page after 2 retries" in str(e.value)
    assert mock_fetch_page.call_count == 3
    assert mock_sleep.call_count == 2  # Should sleep after each failed attempt



# -------------------
# fetch_page (direct)
# -------------------

@patch("history4feed.h4fscripts.h4f.logger")
def test_fetch_page_success(mock_logger, dummy_url):
    session = MagicMock()
    response = MagicMock(spec=Response)
    response.ok = True
    response.content = b"html-content"
    response.headers = {"content-type": "text/html"}
    response.url = dummy_url
    session.get.return_value = response

    content, content_type, final_url = fetch_page(session, dummy_url)

    session.get.assert_called_once_with(dummy_url, headers={})
    assert content == b"html-content"
    assert content_type == "text/html"
    assert final_url == dummy_url

@patch('history4feed.h4fscripts.h4f.fetch_with_scapfly')
def test_fetch_page_uses_scrapfly(mock_fetch, settings, dummy_url):
    api_key = "some value"
    session = MagicMock()
    settings.HISTORY4FEED_SETTINGS = {
        'SCRAPFLY_APIKEY': api_key
    }
    mocked_resp = MagicMock()
    mock_fetch.return_value = [None, mocked_resp]
    result = fetch_page(session, dummy_url)
    mock_fetch.assert_called_once_with(session, dummy_url, {}, api_key)
    assert result == (mocked_resp.content.encode(), mocked_resp.content_type, mocked_resp.url)


@patch("history4feed.h4fscripts.h4f.logger")
def test_fetch_page_not_ok_raises(mock_logger, dummy_url):
    session = MagicMock()
    response = MagicMock()
    response.ok = False
    response.status_code = 500
    response.reason = "Internal Error"
    session.get.return_value = response

    with pytest.raises(history4feedException):
        fetch_page(session, dummy_url)

    session.get.assert_called_once_with(dummy_url, headers={})


@patch("history4feed.h4fscripts.h4f.logger")
@patch("history4feed.h4fscripts.h4f.brotli.decompress", side_effect=Exception("decompress failed"))
def test_fetch_page_brotli_fails(mock_brotli, mock_logger, dummy_url):
    session = MagicMock()
    response = MagicMock()
    response.ok = True
    response.content = b"compressed"
    response.headers = {"content-type": "text/html"}
    response.url = dummy_url
    session.get.return_value = response

    content, ctype, url = fetch_page(session, dummy_url)
    mock_brotli.assert_called_once_with(b"compressed")
    assert content == b"compressed"  # returned as-is due to decompress failure


# ------------------------
# fetch_with_scapfly
# ------------------------

def test_fetch_with_scapfly_success(dummy_url):
    session = MagicMock()
    result_data = {
        "result": {
            "content": "<html></html>",
            "content_type": "text/html",
            "url": dummy_url,
            "status_code": 200,
            "status": "OK"
        }
    }
    response = MagicMock()
    response.status_code = 200
    response.json.return_value = result_data
    session.get.return_value = response

    headers = {"User-Agent": "curl"}
    proxy_apikey = "abc123"

    returned_headers, result = fetch_with_scapfly(session, dummy_url, headers, proxy_apikey)

    assert result.content == "<html></html>"
    assert result.content_type == "text/html"
    assert result.url == dummy_url

    expected_params = {
        "headers[User-Agent]": "curl",
        "key": proxy_apikey,
        "url": dummy_url,
        "country": "us,ca,mx,gb,fr,de,au,at,be,hr,cz,dk,ee,fi,ie,se,es,pt,nl"
    }
    session.get.assert_called_once_with("https://api.scrapfly.io/scrape", params=expected_params)


def test_fetch_with_scapfly_fail_error(dummy_url):
    session = MagicMock()
    response = MagicMock()
    response.status_code = 500
    response.json.return_value = {"result": {}, "message": "Server error"}
    session.get.return_value = response

    with pytest.raises(ScrapflyError):
        fetch_with_scapfly(session, dummy_url, {"User-Agent": "UA"}, "apikey")


def test_fetch_with_scapfly_500_error(dummy_url):
    session = MagicMock()
    response = MagicMock()
    response.status_code = 200
    response.json.return_value = {"result": {"status_code": 500,}, "message": "Server error"}
    session.get.return_value = response

    with pytest.raises(FatalError):
        fetch_with_scapfly(session, dummy_url, {"User-Agent": "UA"}, "apikey")


def test_fetch_with_scapfly_redirect(dummy_url):
    session = MagicMock()
    response = MagicMock()
    response.status_code = 200
    response.json.return_value = {
        "result": {
            "status_code": 302,
            "status": "Redirected",
            "content": "",
            "url": dummy_url,
            "content_type": "text/html"
        }
    }
    session.get.return_value = response

    with pytest.raises(FetchRedirect):
        fetch_with_scapfly(session, dummy_url, {"User-Agent": "UA"}, "apikey")


# ---------------------
# get_full_text
# ---------------------

@patch("history4feed.h4fscripts.h4f.ReadabilityDocument", side_effect=ReadabilityDocument)
@patch("history4feed.h4fscripts.h4f.fetch_page_with_retries")
@patch.object(ReadabilityDocument, 'summary', side_effect=ReadabilityDocument.summary, autospec=True)
def test_get_full_text_success(mock_summary, mock_fetch, mock_readability, dummy_url):
    mock_fetch.return_value = (b"<html>Article</html>", "text/html", dummy_url)

    summary, ctype = get_full_text(dummy_url)

    assert isinstance(summary, str)
    assert ctype == "text/html"
    mock_fetch.assert_called_once_with(dummy_url)
    mock_readability.assert_called_once_with("<html>Article</html>", url=dummy_url)
    mock_summary.assert_called_once()


@patch("history4feed.h4fscripts.h4f.fetch_page_with_retries", side_effect=Exception("failure"))
def test_get_full_text_raises(mock_fetch, dummy_url):
    with pytest.raises(history4feedException) as e:
        get_full_text(dummy_url)
    assert "Error processing fulltext: failure" in str(e.value)


@patch('history4feed.h4fscripts.h4f.fetch_page_with_retries')
@patch('history4feed.h4fscripts.h4f.parse_feed_from_content')
def test_parse_feed_from_url(mock_parse: MagicMock, mock_fetch: MagicMock):
    mock_fetch.return_value = [1, 2, 3]
    url = "https://soem.url/"
    parse_feed_from_url(url)
    mock_fetch.assert_called_once_with(url, retry_count=0)
    mock_parse.assert_called_once_with(1, 3)



# Generated by Django 5.1.6 on 2025-05-02 13:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('history4feed', '0003_alter_feed_description'),
    ]

    operations = [
        migrations.AlterField(
            model_name='fulltextjob',
            name='status',
            field=models.CharField(choices=[('retrieved', 'Retrieved'), ('skipped', 'Skipped'), ('cancelled', 'Cancelled'), ('failed', 'Failed'), ('retrieving', 'Retrieving')], default='retrieving', max_length=15),
        ),
        migrations.AlterField(
            model_name='job',
            name='state',
            field=models.CharField(choices=[('pending', 'Pending'), ('running', 'Running'), ('success', 'Success'), ('cancelled', 'Cancelled'), ('failed', 'Failed')], default='pending', help_text='state of the job', max_length=12),
        ),
    ]



import time
from celery import shared_task, Task as CeleryTask
import celery
from celery.result import ResultSet, AsyncResult
import redis

from history4feed.h4fscripts.sitemap_helpers import fetch_posts_links_with_serper

from ..app import models
from . import h4f, wayback_helpers, logger, exceptions
from datetime import UTC, datetime
from history4feed.app.settings import history4feed_server_settings as settings

from urllib.parse import urlparse
from contextlib import contextmanager
from django.core.cache import cache
from rest_framework.exceptions import APIException, Throttled
from django.db import transaction

LOCK_EXPIRE = 60 * 60

def get_lock_id(feed: models.Feed):
    lock_id = f"feed-lock-{feed.id}"
    logger.debug("using lock id %s", lock_id)
    return lock_id

def queue_lock(feed: models.Feed, job=None):
    lock_value = dict(feed_id=str(feed.id))
    if job:
        lock_value["job_id"] = str(job.id)
        
    status = cache.add(get_lock_id(feed), lock_value, timeout=LOCK_EXPIRE)
    return status


@transaction.atomic()
def new_job(feed: models.Feed, include_remote_blogs):
    job_obj = models.Job.objects.create(
        feed=feed,
        earliest_item_requested=feed.latest_item_pubdate or settings.EARLIEST_SEARCH_DATE,
        latest_item_requested=datetime.now(UTC),
        include_remote_blogs=include_remote_blogs,
    )
    if not queue_lock(feed, job_obj):
        raise Throttled(detail={"message": "A job is already running for this feed", **cache.get(get_lock_id(feed))})

    (start_job.s(job_obj.pk) | retrieve_posts_from_links.s(job_obj.pk) | wait_for_all_with_retry.s() | collect_and_schedule_removal.si(job_obj.pk)).apply_async(countdown=5, link_error=error_handler.s(job_obj.pk))
    return job_obj

def new_patch_posts_job(feed: models.Feed, posts: list[models.Post], include_remote_blogs=True):
    job_obj = models.Job.objects.create(
        feed=feed,
        state=models.JobState.PENDING,
        include_remote_blogs=include_remote_blogs,
    )
    ft_jobs = [models.FulltextJob.objects.create(
        job_id=job_obj.id,
        post_id=post.id,
        link=post.link,
    ) for post in posts]
    chain = celery.chain([retrieve_full_text.si(ft_job.pk) for ft_job in ft_jobs])
    ( start_post_job.si(job_obj.id) | chain | collect_and_schedule_removal.si(job_obj.pk)).apply_async(link_error=error_handler.s(job_obj.pk), countdown=5)
    return job_obj

@shared_task(bind=True, default_retry_delay=10)
def start_post_job(self: CeleryTask, job_id):
    job = models.Job.objects.get(pk=job_id)
    if job.is_cancelled():
        job.info = "job cancelled while in queue"
        job.save(update_fields=['info'])
        return False
    if not queue_lock(job.feed, job):
        return self.retry(max_retries=360)
    job.update_state(models.JobState.RUNNING)
    return True

@shared_task
def start_job(job_id):
    job = models.Job.objects.get(pk=job_id)
    feed = job.feed
    job.update_state(models.JobState.RUNNING)
    try:
        if feed.feed_type == models.FeedType.SEARCH_INDEX:
            return [feed.url]
        return wayback_helpers.get_wayback_urls(feed.url, job.earliest_item_requested, job.latest_item_requested)
    except BaseException as e:
        job.update_state(models.JobState.FAILED)
        job.info = str(e)
        job.save(update_fields=['info'])
        return []

@shared_task(bind=True, default_retry_delay=10)
def wait_for_all_with_retry(self, result_ids):
    if not result_ids:
        return []
    result_set = ResultSet([AsyncResult(task_id) for task_id in result_ids])
    if not result_set.ready():
        return self.retry(max_retries=360)
    return result_ids

@shared_task
def retrieve_posts_from_links(urls, job_id):
    if not urls:
        return []
    full_text_chain = models.Job.objects.get(pk=job_id)
    feed = full_text_chain.feed
    chains = []
    parsed_feed = {}
    job = models.Job.objects.get(id=job_id)
    for index, url in enumerate(urls):
        if job.is_cancelled():
            break
        error = None
        if feed.feed_type == models.FeedType.SEARCH_INDEX:
            posts = retrieve_posts_from_serper(feed, job, url)
        else:
            parsed_feed, posts, error = retrieve_posts_from_url(url, feed, job)
        if error:
            logger.exception(error)
            continue
        if not posts:
            logger.warning('no new post in `%s`', url)
            continue

        full_text_chain = create_fulltexts_task_chain(job_id, posts)
        chains.append(full_text_chain.apply_async())

    if parsed_feed:
        feed.set_description(parsed_feed['description'])
        feed.set_title(parsed_feed['title'])
    feed.freshness = job.run_datetime
    
    feed.save()
    logger.info("====\n"*5)
    return [result.id for result in chains]

def create_fulltexts_task_chain(job_id, posts):
    chain_tasks = []
    for post in posts:
        ftjob_entry = models.FulltextJob.objects.create(
                job_id=job_id,
                post_id=post.id,
                link=post.link,
            )
        chain_tasks.append(retrieve_full_text.si(ftjob_entry.pk))
    return celery.chain(chain_tasks)

def retrieve_posts_from_serper(feed, job, url):
    start_time = feed.freshness or settings.EARLIEST_SEARCH_DATE
    if not start_time.tzinfo:
        start_time = start_time.replace(tzinfo=UTC)
    crawled_posts = fetch_posts_links_with_serper(url, from_time=start_time, to_time=job.run_datetime)
    posts = []
    for post_dict in crawled_posts.values():
        if post := add_post_to_db(feed, job, post_dict):
            posts.append(post)
    return posts

class JobCancelled(Exception):
    pass

@shared_task(bind=True)
def collect_and_schedule_removal(sender, job_id):
    logger.print(f"===> {sender=}, {job_id=} ")
    job = models.Job.objects.get(pk=job_id)
    remove_lock(job)
    if job.state == models.JobState.RUNNING:
        job.update_state(models.JobState.SUCCESS)

def remove_lock(job):
    if cache.delete(get_lock_id(job.feed)):
        logger.debug("lock deleted")
    else:
        logger.debug("Failed to remove lock")

def retrieve_posts_from_url(url, db_feed: models.Feed, job: models.Job):
    back_off_seconds = settings.WAYBACK_SLEEP_SECONDS
    all_posts: list[models.Post] = []
    error = None
    parsed_feed = {}
    for i in range(settings.REQUEST_RETRY_COUNT):
        if i != 0:
            time.sleep(back_off_seconds)
        try:
            if job.is_cancelled():
                raise JobCancelled("job was terminated by user")
            data, content_type, url = h4f.fetch_page_with_retries(url)
            parsed_feed = h4f.parse_feed_from_content(data, url)
            match parsed_feed['feed_type']:
                case models.FeedType.ATOM:
                    posts = h4f.parse_posts_from_atom_feed(url, data)
                case models.FeedType.RSS:
                    posts = h4f.parse_posts_from_rss_feed(url, data)
                case _:
                    raise exceptions.UnknownFeedtypeException("unknown feed type `{}` at {}".format(parsed_feed['feed_type'], url))
            for post_dict in posts.values():
                # make sure that post and feed share the same domain
                post = add_post_to_db(db_feed, job, post_dict)
                if not post:
                    continue
                all_posts.append(post)
            db_feed.save()
            logger.info(f"saved {len(posts)} posts for {url}")
            break
        except ConnectionError as e:
            logger.error(e, exc_info=True)
            error = e
            logger.info(f"job with url {url} ran into an issue {e}, backing off for {back_off_seconds} seconds")
            back_off_seconds *= 1.2
        except BaseException as e:
            logger.error(e, exc_info=True)
            error = e
            break
    return parsed_feed, all_posts, error

def add_post_to_db(db_feed: models.Feed, job: models.Job, post_dict: h4f.PostDict):
    # make sure that post and feed share the same domain
    if job.should_skip_post(post_dict.link):
        models.FulltextJob.objects.create(
                job_id=job.id,
                status=models.FullTextState.SKIPPED,
                link=post_dict.link,
        )
        return None
    categories = post_dict.categories
    del post_dict.categories
    post, created = models.Post.objects.get_or_create(defaults=post_dict.__dict__, feed=db_feed, link=post_dict.link)
    if not created or post.deleted_manually:
        return None

    post.save()
    post.add_categories(categories)
    return post
        
@shared_task()
def retrieve_full_text(ftjob_pk):
    fulltext_job = models.FulltextJob.objects.get(pk=ftjob_pk)
    try:
        if fulltext_job.is_cancelled():
            raise JobCancelled()
        else:
            fulltext_job.post.description, fulltext_job.post.content_type = h4f.get_full_text(fulltext_job.post.link)
            fulltext_job.status = models.FullTextState.RETRIEVED
            fulltext_job.error_str = ""
            fulltext_job.post.is_full_text = True
    except JobCancelled:
        fulltext_job.status = models.FullTextState.CANCELLED
        fulltext_job.error_str = "job cancelled while retrieving fulltext"
    except BaseException as e:
        fulltext_job.error_str = str(e)
        fulltext_job.status = models.FullTextState.FAILED
    fulltext_job.save()
    fulltext_job.post.save()



from celery import signals
@signals.worker_ready.connect
def mark_old_jobs_as_failed(**kwargs):
    models.Job.objects.filter(state__in=[models.JobState.PENDING, models.JobState.RUNNING]).update(state=models.JobState.CANCELLED, info="job cancelled automatically on server startup")

@shared_task
def error_handler(request, exc: Exception, traceback, job_id):
    job = models.Job.objects.get(pk=job_id)
    job.update_state(models.JobState.FAILED)
    job.info = f"job failed: {exc}"
    job.save(update_fields=['info'])
    remove_lock(job)
    logger.error('Job {3} with task_id {0} raised exception: {1!r}\n{2!r}'.format(
          request.id, exc, traceback, job_id))